#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Data pre-processing: build vocabularies and binarize training data.
"""

import argparse
import glob
import os
import random

import soundfile


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "root", metavar="DIR", help="root directory containing flac files to index"
    )
    parser.add_argument(
        "--valid-percent",
        default=0.01,
        type=float,
        metavar="D",
        help="percentage of data to use as validation set (between 0 and 1)",
    )
    parser.add_argument(
        "--dest", default=".", type=str, metavar="DIR", help="output directory"
    )
    parser.add_argument(
        "--ext", default="flac", type=str, metavar="EXT", help="extension to look for"
    )
    parser.add_argument("--seed", default=42, type=int, metavar="N", help="random seed")
    parser.add_argument(
        "--path-must-contain",
        default=None,
        type=str,
        metavar="FRAG",
        help="if set, path must contain this substring for a file to be included in the manifest",
    )
    parser.add_argument(
        "--ignore", default="", type=str, metavar="IGNORE_DIR", help="ignore directory"
    )
    return parser

# ignore_dirs: dict of ignore dirs, such as ['ATEC', 'bq_corpus', 'data', 'tts_audio_8k']
def find_files(root_dir, ext, ignore_dirs):
    list_dirs = os.walk(root_dir)
    search_path = os.path.join(root_dir, "*." + ext)
    file_lists = glob.glob(search_path, recursive=True)
    for root, dirs, _ in list_dirs:
        if os.path.basename(root) in ignore_dirs:
            dirs[:] = []
            ignore_dirs.remove(os.path.basename(root))
        else:
            for file_dir in dirs:
                search_path = os.path.join(root, file_dir, "*." + ext)
                print("Search: ", search_path)
                file_lists += (glob.glob(search_path, recursive=True))
    return file_lists

def main(args):
    assert args.valid_percent >= 0 and args.valid_percent <= 1.0
    ignore_dirs = args.ignore.strip().split(',')
    print("Ignore dirs: ", ignore_dirs)
    dir_path = os.path.realpath(args.root)
    file_lists = find_files(dir_path, args.ext, ignore_dirs)
    rand = random.Random(args.seed)
    durations = 0.0
    with open(os.path.join(args.dest, "train.tsv"), "w") as train_f, open(
        os.path.join(args.dest, "valid.tsv"), "w"
    ) as valid_f:
        print(dir_path, file=train_f)
        print(dir_path, file=valid_f)

        for fname in file_lists:
            file_path = os.path.realpath(fname)

            if args.path_must_contain and args.path_must_contain not in file_path:
                continue

            frames = soundfile.info(fname).frames
            durations += soundfile.info(fname).duration
            dest = train_f if rand.random() > args.valid_percent else valid_f
            print(
                "{}\t{}".format(os.path.relpath(file_path, dir_path), frames), file=dest
            )
    
    print("Finish all files with duration: ", durations)
    hours, rest = divmod(durations, 3600)
    minutes, seconds = divmod(rest, 60)
    if hours >= 1:
        duration_str = "{}:{}:{} h".format(hours, minutes, seconds)
    elif minutes >= 1:
        duration_str = "{}:{} min".format(minutes, seconds)
    else:
        duration_str = "{} s".format(seconds)
    print('Total times: ', duration_str)

if __name__ == "__main__":
    parser = get_parser()
    args = parser.parse_args()
    main(args)
